__version__ = '0.0.1'

import argparse
import configparser
from pathlib import Path
import sys

from habutax import enum
from habutax import fields
from habutax import form
from habutax import forms
from habutax import inputs
from habutax import pdf_fields
from habutax import pdf_filler
from habutax import solver
from habutax import values

def prompt_input(missing, needed_by):
    """
    Given `missing`, an Input which the solver identified as being needed but
    not supplied, prompt the user to supply it. `needed_by` is a list of the
    Field objects which need the input.
    """

    # Prompt the user to input missing input value
    prompt = f'\n----[ {missing.name()} ]----'
    prompt += '-' * max(0, (80 - len(prompt))) + '\n'

    # Provide context for why this question is being asked
    prompt += 'Additional input is needed by:\n'
    duplicates = {}
    for f in needed_by:
        form_desc = f.form().full_description()
        field_basename = f.base_name()
        instance = f'Instance \'{f.form().instance()}\' of ' if f.form().instance() else ''
        potential_prompt = f' * {instance}{form_desc}, line \'{field_basename}\'\n'
        if potential_prompt not in duplicates:
            duplicates[potential_prompt] = True
            prompt += potential_prompt

    # Include any help from the input specification about the question being
    # asked or the expected format of the response
    prompt += '\n-> ' + missing.help() + '\n\n'
    format_suggestion = missing.format_suggestion()
    if len(format_suggestion) > 0:
        prompt += format_suggestion + '\n'

    prompt += f'(Ctrl-C to abort): '

    value = None

    while value is None or not missing.valid(value):
        try:
            if value is not None:
                prompt = "Invalid input, try again?: "
            value = input(prompt)
        except KeyboardInterrupt:
            return (None, False)

    return (value, True)

def solve(args):
    if args.writeback_input:
        Path(args.input_file).touch() # Ensure the input file exists (this allows writing back without the user having to manually touch it first)

    input_store = inputs.InputStore(args.input_file)
    prompt_fn = prompt_input if args.prompt_missing else None
    s = solver.Solver(input_store, forms.available_forms[args.year], prompt=prompt_fn)
    try:
        successful = s.solve(args.forms)
        solution = s.solution()
    except Exception as e:
        raise e
    finally:
        # Ensure that even if an exception happens, any output the user already
        # entered is saved as they requested
        if args.writeback_input:
            input_store.write(args.input_file)

    # Attach tax year to solution
    solution['habutax'] = {
        'tax_year': args.year,
        'version': __version__,
    }

    if successful:
        print("\nSuccessfully solved!")
    else:
        print("\nFailed to solve, because...")

        unimplemented_fields = s.unimplemented_fields()
        unmet_input_dependencies = s.unmet_input_dependencies()
        unmet_field_dependencies = s.unmet_field_dependencies()

        if len(unimplemented_fields) > 0:
            print("\nThe following fields encountered unimplemented behavior:")
            for unimplemented in unimplemented_fields:
                print(f'- {unimplemented}')
        if len(unmet_input_dependencies) > 0:
            print("\nThe following inputs were needed but not supplied:")
            for dependency, dependents in unmet_input_dependencies.items():
                print(f'{dependency} (needed by: {", ".join(dependents)})')
        if len(unmet_field_dependencies) > 0:
            print("\nThe following fields were needed but unable to be produced (likely due to unsupplied inputs or unimplemented behavior above):")
            for dependency, dependents in unmet_field_dependencies.items():
                print(f'{dependency} (needed by: {", ".join(dependents)})')

    if not args.solution:
        class StringWriter(object):
            def __init__(self):
                self.output = ""
            def write(self, written):
                self.output += written
            def __str__(self):
                return self.output
        string_writer = StringWriter()
        solution.write(string_writer)

        print(string_writer)
    else:
        with open(args.solution, 'w') as outfile:
            solution.write(outfile)
        print(f'\nSolver results written to {args.solution}')

def fill_pdfs(args):
    solution = configparser.ConfigParser()
    with open(args.solution) as solution_file:
        solution.read_file(solution_file)
    tax_year = solution.getint('habutax', 'tax_year')

    # Remove this 'special' section so the PDF filler doesn't interpret it as
    # form data to be filled
    solution.remove_section('habutax')

    p = pdf_filler.PDFFiller(solution, forms.available_forms[tax_year], args.output, flatten=args.flatten)
    p.fill()

def list_forms(args):
    header_printed = False
    if args.contains is not None:
        search_string = args.contains.lower()

    width = max([len(f.form_name) for f in forms.available_forms[args.year]])
    format_str ="{:>{width}} | {:12} | {}"

    for form in forms.available_forms[args.year]:
        name = form.form_name

        description = f'{form.description}: {form.long_description}'
        search_match = args.contains is None or search_string in name.lower() or search_string in description.lower()
        jurisdiction_match = args.jurisdiction is None or args.jurisdiction.lower() == form.jurisdiction.name.lower()
        if search_match and jurisdiction_match:
            if not header_printed:
                header_printed = True
                if args.contains:
                    print("Matching forms:")
                else:
                    print("Form list:")
                print(format_str.format("name", "jurisdiction", "description", width=width))
                print(format_str.format("-"*width, "-"*12, "-"*width, width=width))
            print(format_str.format(name, form.jurisdiction.name, description, width=width))

    if not header_printed and (args.contains or args.jurisdiction):
        print("No forms matched your parameters")

def list_form_inputs(args):
    form_name, form_instance = form.name_and_instance(args.form)
    forms_map = {f.form_name : f for f in forms.available_forms[args.year]}

    if form_name not in forms_map:
        print(f'Cannot find form "{args.form}". The `list-forms` sub-command will list the valid form names.')
        sys.exit(1)

    form_class = forms_map[form_name]
    if hasattr(form_class, 'valid_instances'):
        if form_instance not in form_class.valid_instances:
            instances = ", ".join([f'"{i}"' for i in form_class.valid_instances])
            print(f'Form "{form_name}" requires a form instance be specified and that it be one of {instances}. For example, try "{form_name}:{form_class.valid_instances[0]}".')
            sys.exit(1)

    f = form_class(instance=form_instance)
    print(f'[{f.name()}]')
    print(f'# {f.full_description()}')

    input_map = {i.name(): i for i in f.inputs()}
    sorted_inputs = sorted(input_map.keys(), key=solver.sort_keys)

    for input_name in sorted_inputs:
        form_input = input_map[input_name]
        comment = '# ' + form_input.help() + '\n'
        format_suggestion = form_input.format_suggestion()
        if len(format_suggestion) > 0:
            comment += format_suggestion
        print('\n' + comment.replace('\n', '\n# '))
        print(f'#{form_input.base_name()} =')

def main():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(required=True, help='sub-command help')
    default_year = 2021

    # Solver argument setup
    solve_parser = subparsers.add_parser('solve', help='Solve taxes using HabuTax')
    solve_parser.add_argument('input_file', type=str, help='The file containing your input for the tax forms you are calculating.')
    solve_parser.add_argument('--year', type=int, default=default_year, help=f'The tax year to use (default: {default_year})')
    solve_parser.add_argument('--form', dest='forms', action='append', help='Which form(s) you want to calculate')
    solve_parser.add_argument('--prompt-missing', action='store_true', default=False, help='Interactively prompt for any missing input')
    solve_parser.add_argument('--writeback-input', action='store_true', default=False, help='Write any interactively-supplied input back to the config file when done (loses any comments/formatting present in file)')
    solve_parser.add_argument('--solution', type=str, default=None, help='Output text file for the results of the tax solver (defaults to stdout)')
    solve_parser.set_defaults(func=solve)

    # fill-pdfs argument setup
    fill_pdfs_parser = subparsers.add_parser('fill-pdfs', help='Fill PDFs using a solution previously calculated using HabuTax')
    fill_pdfs_parser.add_argument('--no-flatten', dest='flatten', action='store_false', default=True, help='Do not "flatten" the PDF after filling, so its fields can still be edited')
    fill_pdfs_parser.add_argument('solution', type=str, help='The file containing the solution of the tax forms you want to generate PDFs of.')
    fill_pdfs_parser.add_argument('output', type=str, help='Path where you want to write the generated PDF file')
    fill_pdfs_parser.set_defaults(func=fill_pdfs)

    # list-forms argument setup
    list_forms_parser = subparsers.add_parser('list-forms', help='List all available forms, optionally restricting to those containing a string or within a jurisdiction')
    list_forms_parser.add_argument('--year', type=int, default=default_year, help=f'The tax year to use (default: {default_year})')
    list_forms_parser.add_argument('--contains', type=str, default=None, help='A string to search for (only print forms containing this string)')
    list_forms_parser.add_argument('--jurisdiction', type=str, default=None, help='Only display forms from this tax jurisdiction ("US" federal, "NC" state, etc.)')
    list_forms_parser.set_defaults(func=list_forms)

    # list-form-inputs argument setup
    list_form_inputs_parser = subparsers.add_parser('list-form-inputs', help='List all available inputs for a given form, in a format suitable for editing and passing as input to HabuTax.')
    list_form_inputs_parser.add_argument('form', type=str, help='The short name of the form for which you want to display the inputs')
    list_form_inputs_parser.add_argument('--year', type=int, default=default_year, help=f'The tax year to use (default: {default_year})')
    list_form_inputs_parser.set_defaults(func=list_form_inputs)

    args = parser.parse_args()
    args.func(args)



import unittest

from .. import context

import habutax.forms.ty2021 as ty2021
from habutax.inputs import InputStore
from habutax.solver import Solver

class Form1040SimpleTestCase(unittest.TestCase):
    def setUp(self):
        self.inputs = InputStore('tests/ty2021/fixtures/one_w2.habutax')
        self.solver = Solver(self.inputs, ty2021.available_forms)

    def assertDollarsEqual(self, solution, field, expected):
        form, field = field.split('.')
        calculated = float(solution[form][field])
        self.assertAlmostEqual(calculated, expected, places=2)

    def test_simple_1040(self):
        solved = self.solver.solve(['1040'])
        solution = self.solver.solution()

        self.assertTrue(solved)

        self.assertIn('1040', solution)
        self.assertIn('w-2:0', solution)

        self.assertDollarsEqual(solution, '1040.1', 100000)
        self.assertDollarsEqual(solution, '1040.12c', 12550 + 300)
        self.assertDollarsEqual(solution, '1040.15', 100000 - (12550 + 300))
        self.assertDollarsEqual(solution, '1040.16', 14943)
        self.assertDollarsEqual(solution, '1040.25a', 16551.9)
        self.assertDollarsEqual(solution, '1040.25d', 16551.9)
        self.assertDollarsEqual(solution, '1040.33', 16551.9)
        self.assertDollarsEqual(solution, '1040.34', 1608.9)
        self.assertDollarsEqual(solution, '1040.35a', 1608.9)
        self.assertDollarsEqual(solution, '1040.37', 0)
        self.assertDollarsEqual(solution, '1040.38', 0)



import unittest

from .. import context

from habutax.forms.ty2021.f1040_figure_tax import figure_tax
from habutax import enum

class FigureTaxTestCase(unittest.TestCase):
    def setUp(self):
        self.FILING_STATUS = enum.make('1040 Filing Status', {
            'Single': "single, unmarried, or legally separated",
            'MarriedFilingJointly': "married and filing a joint return",
            'MarriedFilingSeparately': "married and file a separate return",
            'HeadOfHousehold': "unmarried and provide a home for certain other person",
            'QualifyingWidowWidower': "generally filed if your spouse died in 2019 or 2020 and you didn't remarry before the end of 2021 and you have a child or stepchild whom you can claim as a dependent (see Form 1040 instructions for more)"})

    def test_figure_2021_1040_tax(self):
        scenarios = (
            (75047.08, self.FILING_STATUS.Single, 12254),
            (98325.84, self.FILING_STATUS.MarriedFilingJointly, 13129),
            (99999.99, self.FILING_STATUS.HeadOfHousehold, 16563.00),
            (100000.00, self.FILING_STATUS.HeadOfHousehold, 16569.00),
            (588939, self.FILING_STATUS.QualifyingWidowWidower, 155217.15),
            (588939, self.FILING_STATUS.MarriedFilingJointly, 155217.15),
            (1482.40, self.FILING_STATUS.MarriedFilingSeparately, 149),
            (8182983.94, self.FILING_STATUS.MarriedFilingSeparately, 2995965.31),
        )
        for taxable, status, expected_tax in scenarios:
            tax = figure_tax(taxable, status, self.FILING_STATUS)
            self.assertAlmostEqual(expected_tax, tax, places=2)



from enum import Enum

# Mixin which changes the string formatting to only include the name
class StringyEnum(object):
    def __str__(self):
        return self.name

def make(name, options):
    assert(type(options) == dict)
    for k, v in options.items():
        assert(isinstance(k, str)) 
        assert(isinstance(v, str)) 
    return Enum(name, options, type=StringyEnum)

us_states = make('US States', {
	'AL': 'Alabama',
	'AK': 'Alaska',
	'AZ': 'Arizona',
	'AR': 'Arkansas',
	'CA': 'California',
	'CO': 'Colorado',
	'CT': 'Connecticut',
	'DE': 'Delaware',
	'FL': 'Florida',
	'GA': 'Georgia',
	'HI': 'Hawaii',
	'ID': 'Idaho',
	'IL': 'Illinois',
	'IN': 'Indiana',
	'IA': 'Iowa',
	'KS': 'Kansas',
	'KY': 'Kentucky',
	'LA': 'Louisiana',
	'ME': 'Maine',
	'MD': 'Maryland',
	'MA': 'Massachusetts',
	'MI': 'Michigan',
	'MN': 'Minnesota',
	'MS': 'Mississippi',
	'MO': 'Missouri',
	'MT': 'Montana',
	'NE': 'Nebraska',
	'NV': 'Nevada',
	'NH': 'New Hampshire',
	'NJ': 'New Jersey',
	'NM': 'New Mexico',
	'NY': 'New York',
	'NC': 'North Carolina',
	'ND': 'North Dakota',
	'OH': 'Ohio',
	'OK': 'Oklahoma',
	'OR': 'Oregon',
	'PA': 'Pennsylvania',
	'RI': 'Rhode Island',
	'SC': 'South Carolina',
	'SD': 'South Dakota',
	'TN': 'Tennessee',
	'TX': 'Texas',
	'UT': 'Utah',
	'VT': 'Vermont',
	'VA': 'Virginia',
	'WA': 'Washington',
	'WV': 'West Virginia',
	'WI': 'Wisconsin',
	'WY': 'Wyoming'}
)

taxpayer_or_spouse = make('Taxpayer or Spouse', {
	'taxpayer': 'This belongs to the taxpayer',
	'spouse': 'This belongs to the taxpayer\'s spouse (if filing a joint return)'
})

taxpayer_spouse_or_both = make('Taxpayer, Spouse, or Both', {
	'taxpayer': 'This belongs to the taxpayer',
	'spouse': 'This belongs to the taxpayer\'s spouse (if filing a joint return)',
	'both': 'This belongs to both the taxpayer and their spouse (if filing a joint return)'
})



import sys

from habutax import fields
from habutax import form
from habutax import inputs
from habutax import values

class DependencyTracker(object):
    def __init__(self):
        # map names of unmet dependencies to a list of their outstanding
        # dependents
        self._unmet = {}

        # A list of the names of newly-satisfied dependencies
        self._met = []

    def add_unmet(self, dependency_name, dependent):
        """Add an unmet dependency for a particular dependent"""
        if dependency_name not in self._unmet:
            self._unmet[dependency_name] = [dependent]
        else:
            self._unmet[dependency_name].append(dependent)

    def has_met(self):
        """Return True if there are met dependencies which have not been
        removed by calling met_dependents()""" 
        return len(self._met) > 0

    def has_unmet(self):
        """Returns True if there are still unmet dependencies (does not count
        met dependencies which have not yet been removed via a call to
        met_dependents()"""
        for dependency, dependents in self._unmet.items():
            if len(dependents) > 0 and dependency not in self._met:
                return True
        return False

    def meet(self, dependency_name):
        """Mark a depdendency as having been satisfied (allows any dependents
        to be released in met_dependents()"""
        self._met.append(dependency_name)

    def unmet_dependencies(self):
        """Returns the names of the unmet dependencies"""
        return list(self._unmet.keys())

    def unmet_dependents(self, dependency):
        """Return the list of objects of the dependents waiting on an unmet dependency"""
        return self._unmet[dependency]

    def met_dependents(self):
        """Generator returning the objects representing all dependents whose
        dependencies have been marked as satisfied. Note that this also removes
        the returned dependency, so if a further dependency is later
        discovered, this object must be re-added using add_unmet()"""
        while len(self._met) > 0:
            met = self._met[0]
            # check if there are any dependents waiting on this satisfied dependency
            if met in self._unmet:
                dependent = self._unmet[met].pop()

                if len(self._unmet[met]) == 0:
                    del self._unmet[met]
                    self._met.pop(0)

                yield dependent
            else:
                self._met.pop(0)
                continue

def _sort_keys(key):
    last_numeric = False
    last_alpha = False
    current = ''
    keys = []

    for c in key:
        this_numeric = c in '01234567890'
        this_alpha = c.isalpha()
        if (last_numeric and not this_numeric) or \
                (last_alpha and not this_alpha):
            if len(current) > 0:
                if last_numeric:
                    keys.append((False, int(current)))
                else:
                    keys.append((True, current))
                current = ''

        if this_numeric or this_alpha:
            current += c
        else:
            current = ''

        last_numeric = this_numeric
        last_alpha = this_alpha

    # Output the last partial sub-key in the string
    if len(current) > 0:
        if last_numeric:
            keys.append((False, int(current)))
        else:
            keys.append((True, current))

    return keys

def sort_keys(key):
    if isinstance(key, fields.Field) or isinstance(key, inputs.Input):
        key = key.name()

    if '.' in key:
        form, key = key.split('.')
    else:
        form = ''
    return (_sort_keys(form), _sort_keys(key))

class Solver(object):
    def __init__(self, input_config, form_list, prompt=None):
        self._prompt = prompt
        self._refused_input = self._prompt is None

        # Create a map to easily look up the available forms by name
        self._form_map = {f.form_name: f for f in form_list}

        # Map all available input and field names to their objects
        self._input_map = {}
        self._field_map = {}

        # Input values (also modified with any values the user specifies
        # interactively)
        self._i = input_config

        # Current state of solver, including form instances, values calculated,
        # any unmet field/input dependencies
        self.forms = {}
        self._v = values.ValueStore()
        self._unattempted_fields = []
        self._unimplemented_fields = []
        self._solving_fields = set()
        self._field_dependencies = DependencyTracker()
        self._input_dependencies = DependencyTracker()

        self._done_solving = False # Set to True if/when done solving
        self._solved = False       # Set to True if/when successfully solved

    def _add_unattempted(self, unattempted):
        if isinstance(unattempted, list):
            self._unattempted_fields.extend(unattempted)
        else:
            self._unattempted_fields.append(unattempted)
        self._unattempted_fields.sort(key=sort_keys)

    def _add_form(self, form_name):
        form_name, form_instance = form.name_and_instance(form_name)

        if form_name not in self._form_map:
            raise NotImplementedError(f'Form {form_name} is not supported.')

        new_form = self._form_map[form_name](solver=self, instance=form_instance)
        self.forms[new_form.name()] = new_form

        # Add new inputs to our internal map of names to input objects, update
        # the input mapper so it understands how to read these inputs
        for i in new_form.inputs():
            assert(i not in self._input_map)
            self._input_map[i.name()] = i
        self._i.update_input_spec(self._input_map)

        for f in new_form.fields():
            assert(f not in self._field_map)
            self._field_map[f.name()] = f

        self._add_unattempted(new_form.required_fields())
        self._solving_fields |= set([f.name() for f in new_form.required_fields()])

    def _attempt_input(self, input_name, needed_by):
        missing = self._input_map[input_name]

        value, supplied = self._prompt(missing, needed_by)

        if supplied:
            assert(missing.valid(value))
            self._i[missing.name()] = value
            self._input_dependencies.meet(missing.name())
        else:
            self._refused_input = True
        return supplied

    def _attempt_field(self, field):
        try:
            form_inputs = form.FormAccessor(self._i, field.form())
            form_values = form.FormAccessor(self._v, field.form())
            self._v[field.name()] = field.value(form_inputs, form_values)
            self._field_dependencies.meet(field.name())
        except values.UnmetDependency as ud:
            # If this field is not already in the fields the solver is
            # attempting to solve, add it
            if ud.dependency not in self._solving_fields:
                # If this field is not even in the list of forms the solver is
                # aware of, it must be in another form - find that form
                if ud.dependency not in self._field_map:
                    form_name, field_name = ud.dependency.split('.')
                    self._add_form(form_name)
                assert(ud.dependency in self._field_map)
                self._add_unattempted(self._field_map[ud.dependency])
                self._solving_fields.add(ud.dependency)

            self._field_dependencies.add_unmet(ud.dependency, field)
        except inputs.MissingInput as mi:
            self._input_dependencies.add_unmet(mi.input_name, field)
        except fields.FieldNotImplemented as fni:
            self._unimplemented_fields.append(fni.field_name)

    def solve(self, form_names):
        for form_name in form_names:
            self._add_form(form_name)

        while len(self._unattempted_fields) > 0 \
                or self._input_dependencies.has_met() \
                or (self._input_dependencies.has_unmet() and not self._refused_input) \
                or self._field_dependencies.has_met():

            while len(self._unattempted_fields) > 0:
                self._attempt_field(self._unattempted_fields.pop())
            for field in sorted(self._field_dependencies.met_dependents(), key=sort_keys):
                self._attempt_field(field)
            if not self._refused_input:
                for input_name in sorted(self._input_dependencies.unmet_dependencies(), key=sort_keys):
                    needed_by = self._input_dependencies.unmet_dependents(input_name)
                    self._attempt_input(input_name, needed_by)
                    if self._refused_input:
                        break
            # Must be list() because _attempt_field modifies
            # input_dependencies, and we don't want to get stuck in a loop
            # waiting for an input we can't provide without polling user for it
            for field in list(self._input_dependencies.met_dependents()):
                self._attempt_field(field)

        assert(len(self._unattempted_fields) == 0)
        assert(not self._input_dependencies.has_met())
        assert(not self._field_dependencies.has_met())

        self._done_solving = True

        if not self._field_dependencies.has_unmet() \
                and not self._input_dependencies.has_unmet() \
                and len(self._unimplemented_fields) == 0:
            self._solved = True

        return self._solved

    def solution(self):
        """Return a ConfigParser object representing the portions of the
        requested forms which were successfully solved"""
        assert(self._done_solving)
        return self._v.to_config(self._field_map)

    def unimplemented_fields(self):
        """Return a list of the unimplemented field names after a call to
        solve()"""
        assert(self._done_solving)
        return self._unimplemented_fields

    def _unmet_dependencies(self, dependency_tracker):
        unmet_dependencies = {}
        for dep in dependency_tracker.unmet_dependencies():
            dependents = [f.name() for f in dependency_tracker.unmet_dependents(dep)]
            unmet_dependencies[dep] = dependents
        return unmet_dependencies

    def unmet_input_dependencies(self):
        """Return a dict mapping names of any unmet input dependencies to a
        list of the names of their (discovered) dependent fields after a call
        to solve()"""
        assert(self._done_solving)
        return self._unmet_dependencies(self._input_dependencies)

    def unmet_field_dependencies(self):
        """Return a dict mapping names of any unmet field dependencies to a
        list of the names of their (discovered) dependents after a call to
        solve()"""
        assert(self._done_solving)
        return self._unmet_dependencies(self._field_dependencies)



import os
import tempfile
import subprocess

from habutax import fields
from habutax import inputs
from habutax import values

fdf_header = """%FDF-1.2
%,,oe"
1 0 obj
<< /FDF << /Fields ["""

fdf_footer = """
] >> >>
endobj
trailer
<< /Root 1 0 R >>
%%EOF;
"""

class PDFFiller(object):
    def __init__(self, solution, available_forms, outfilename, flatten=True):
        self._solution = solution
        self._form_map = {f.form_name: f for f in available_forms}
        self._output_filename = outfilename
        self._pdftk = 'pdftk'
        self._flatten = flatten

        # Instances of Forms in the solution, and fields belonging to those
        # forms
        self.forms = []
        self._field_map = {}

        # Values read from solution file
        self._values = values.ValueStore()

    def _read_form_fields(self, form_name):
        # Read values in as their types from solution file's string
        # representation
        for field_name in self._solution[form_name]:
            full_name = f'{form_name}.{field_name}'
            string = self._solution[form_name][field_name]
            field = self._field_map[full_name]
            self._values[full_name] = field.from_string(string)

    def _add_form(self, full_form_name):
        split_form_name = full_form_name.split(':')
        form_instance = None
        if len(split_form_name) == 2:
            form_name = split_form_name[0]
            form_instance = split_form_name[1]
        elif len(split_form_name) != 1:
            raise RuntimeError(f'Unexpected form name: {form_name} (expected 0 or 1 colons)')
        form_name = split_form_name[0]

        if form_name not in self._form_map:
            raise NotImplementedError(f'Form {form_name} is not supported.')

        form = self._form_map[form_name](instance=form_instance)
        self.forms.append(form)

        for f in form.fields():
            assert(f not in self._field_map)
            self._field_map[f.name()] = f

        self._read_form_fields(full_form_name)

    def _create_fdf(self, data, filename):
        lines = []
        for k, v in data.items():
            lines.append(f'<< /T ({k}) /V ({v}) >>')
        with open(filename, 'w') as fdf:
            fdf.write(fdf_header)
            fdf.write("\n".join(lines))
            fdf.write(fdf_footer)

    def _fill_form(self, form, pdf_filename):
        assert form.pdf_file()
        assert len(form.pdf_fields()) > 0
        required_fields = {f.name() for f in form.required_fields()}

        fdf_map = {}
        for pdf_field in form.pdf_fields():
            field_name = pdf_field.field_name
            if "." not in field_name:
                field_name = f'{form.name()}.{field_name}'

            if field_name not in self._field_map:
                raise RuntimeError(f'PDF filler found a field absent from the specification of any supplied form: {field_name} in form {form.name()}')
            try:
                value = self._values[field_name]
                field = self._field_map[field_name]
                string_value = pdf_field.value(value, field)
            except values.UnmetDependency:
                assert field_name not in required_fields
                string_value = ""
            fdf_map[pdf_field.pdf_field_name] = string_value

        fdf_filename = f'{pdf_filename}.fdf'
        self._create_fdf(fdf_map, fdf_filename)

        cmd = [self._pdftk, form.pdf_file(), 'fill_form', fdf_filename, 'output', pdf_filename]
        if self._flatten:
            cmd.append('flatten')
        res = subprocess.run(cmd, check=True)

    def fill(self):
        for form_name in self._solution:
            # Ignore this artifact of ConfigParser
            if form_name == 'DEFAULT':
                continue
            self._add_form(form_name)

        # Filter by forms which need to be filed, and sort the remaining forms
        # by jurisdiction and then sequence number so that they are output in
        # the correct order
        filling_forms = [f for f in self.forms if f.needs_filing(self._values)]
        filling_forms.sort(key=lambda f: (f.jurisdiction, f.sequence_no))

        pdfs = []
        with tempfile.TemporaryDirectory() as tmpdirname:
            for form in filling_forms:
                if form.needs_filing(self._values):
                    pdf_filename = os.path.join(tmpdirname, f'{form.name()}.pdf')
                    self._fill_form(form, pdf_filename)
                    pdfs.append(pdf_filename)

            cmd = [self._pdftk]
            cmd.extend(pdfs)
            cmd.extend(['cat', 'output', self._output_filename])
            res = subprocess.run(cmd, check=True)



from types import MethodType

class PDFValueTooLong(Exception):
    def __init__(self, pdf_field_name, field_name, max_length):
        self.message = f'Value from internal field {field_name} exceeded the max length of PDF field {pdf_field_name} (which is {max_length}) and could not be written'
        self.pdf_field_name = pdf_field_name
        self.field_name = field_name
        self.max_length = max_length
        super().__init__(self.message)

class PDFInvalidChoiceValue(Exception):
    def __init__(self, pdf_field_name, field_name, choice):
        self.message = f'Value from internal field {field_name} does not map to the valid PDF field {pdf_field_name} choices (choice was {choice}) and could not be written'
        self.pdf_field_name = pdf_field_name
        self.field_name = field_name
        self.choice = choice
        super().__init__(self.message)

class PDFField(object):
    def __init__(self, pdf_field_name, field_name, value_fn=None):
        self.pdf_field_name = pdf_field_name
        self.field_name = field_name
        if value_fn is not None:
            self._value_fn = MethodType(value_fn, self)
        else:
            self._value_fn = None

    def value(self, value, field_obj):
        if self._value_fn is not None:
            return self._value_fn(value, field_obj)
        else:
            return field_obj.to_string(value)

class TextPDFField(PDFField):
    def __init__(self, name, value, max_length=None, value_fn=None):
        self.max_length = None
        super().__init__(name, value, value_fn=value_fn)

    def value(self, value, field_obj):
        value = super().value(value, field_obj)
        if self.max_length is not None and len(value) > self.max_length:
            raise PDFValueTooLong(self.pdf_field_name, self.field_name, self.max_length)
        return value

class ButtonPDFField(PDFField):
    def __init__(self, name, value, true_value, value_fn=None):
        self._true_value = true_value # The value to return when this is true
        super().__init__(name, value, value_fn=value_fn)

    def value(self, value, field_obj):
        if self._value_fn is not None:
            value = self._value_fn(value, field_obj)
        if value:
            return self._true_value
        else:
            return 'Off'

class OptionlessButtonPDFField(PDFField):
    def __init__(self, name, value, value_fn=None):
        super().__init__(name, value, value_fn=value_fn)

    def value(self, value, field_obj):
        """Should never be called - so far these are not fields which have needed output"""
        raise NotImplementedError()

class ChoicePDFField(PDFField):
    def __init__(self, name, value, choices, value_fn=None):
        self._choices = choices
        super().__init__(name, value, value_fn=value_fn)

    def value(self, value, field_obj):
        value = super().value(value, field_obj)
        if value not in self._choices:
            raise PDFInvalidChoiceValue(self.pdf_field_name, self.field_name, value)
        return value
